﻿//
//  Win32NativeUiInterface_cpp
//  scummvm
//
//
//

#include "backends/platform/sdl/win32/Win32NativeUiInterface.h"
#include "common/textconsole.h"
#include "backends/platform/sdl/sdl-sys.h"
#include "enhanced/GameUIManager.h"

#include <Windows.h>

// Trying to avoid calling back from C# to C++,
// because this will require a new wrapper project (otherwise circular refereneces appear),
// so all NativeUiInterface methods should be called directly from this class

wchar_t* Win32NativeUiInterface::CreateWCharString(wstring string)
{	
	//setlocale(LC_ALL, "en_US.UTF-8");
	wchar_t* newString = new wchar_t[string.size() + 1 /*for the case of zero-length strings*/];	
	wcscpy(newString, string.c_str());
	return newString;
}

vector<const wchar_t*>* Win32NativeUiInterface::WStringToWCharVector(vector<wstring>& original)
{
	//setlocale(LC_ALL, "en_US.UTF-8");
	vector<const wchar_t*>* result = new vector<const wchar_t*>();

	for (vector<wstring>::iterator iterator = original.begin(); iterator != original.end(); iterator++)
		result->push_back(CreateWCharString(*iterator));

	return result;
}

void Win32NativeUiInterface::DeleteWCharVector(vector<const wchar_t*>* wcharVector)
{
	for (vector<const wchar_t*>::iterator iterator = wcharVector->begin(); iterator != wcharVector->end(); iterator++)
		delete (*iterator);
	delete wcharVector;
}

Win32NativeUiInterface::Win32NativeUiInterface()
{
	_wrapper = new UnmanagedWrapper();
	_localizedStrings = NULL;
	/*_backToMenu = false;
	_isLoadingGameFromMainMenu = false;
	_isStartingNewGame = false;
	_level = 0;*/
	_showingMenu = false;

	_mutexHandle = CreateMutex(NULL, TRUE, "Global\\fec36f40-67ae-4d0e-90eb-305893df9ccf");
	_secondInstance = GetLastError() == ERROR_ALREADY_EXISTS;
}

#define APP_STORE_GOG

const wchar_t* GetPromotionUrl(int gameToPromote)
{
	const wchar_t* urlToPromote = NULL;

#if defined(APP_STORE_GOG)
	switch (gameToPromote) {
	case GAME_TYPE_IHNM:
		urlToPromote = L"https://www.gog.com/game/i_have_no_mouth_and_i_must_scream";
		break;
	case GAME_TYPE_FOTAQ:
		urlToPromote = L"https://www.gog.com/game/flight_of_the_amazon_queen";
		break;
	case GAME_TYPE_SIMON1:
		urlToPromote = L"https://www.gog.com/game/simon_the_sorcerer";
		break;
	case GAME_TYPE_SIMON2:
		urlToPromote = L"https://www.gog.com/game/simon_the_sorcerer_2";
		break;
	case GAME_TYPE_T7G:
		urlToPromote = L"https://www.gog.com/game/the_7th_guest_25th_anniversary_edition";
		break;
	case GAME_TYPE_TONYT:
		urlToPromote = L"https://www.gog.com/game/tony_tough";
		break;
	case GAME_TYPE_ITE:
		urlToPromote = L"https://www.gog.com/game/inherit_the_earth";
		break;
	};
#elif defined(APP_STORE_STEAM)
	switch (gameToPromote) {
	case GAME_TYPE_IHNM:
		urlToPromote = L"http://store.steampowered.com/app/252710";
		break;
	case GAME_TYPE_FOTAQ:
		urlToPromote = L"http://store.steampowered.com/app/252710";
		break;
	case GAME_TYPE_SIMON1:
		urlToPromote = L"http://store.steampowered.com/app/674150";
		break;
	case GAME_TYPE_SIMON2:
		urlToPromote = L"http://store.steampowered.com/app/678760";
		break;
	case GAME_TYPE_T7G:
		urlToPromote = L"https://store.steampowered.com/app/1044340";
		break;
	case GAME_TYPE_TONYT:
		urlToPromote = L"http://store.steampowered.com/app/252710";
		break;
	case GAME_TYPE_ITE:
		urlToPromote = L"http://store.steampowered.com/app/252710";
		break;
	};
#endif

	return urlToPromote;
}

void Win32NativeUiInterface::LoadStrings()
{
	//setlocale(LC_ALL, "en_US.UTF-8");
	_localizedStrings = new vector<const wchar_t*>();
	for (int i = (int)StringId::app_desc; i <= (int)StringId::sfx_volume; i++)
	{
		wstring ws = Strings::getString(static_cast<StringId>(i));
		wchar_t* wc = CreateWCharString(ws);
		_localizedStrings->push_back(wc);
	}

	_wrapper->UpperCase = !hasFontwithLowercase();
	int gameToPromote = getCurrentCrossPromotionGame();
	
	_wrapper->InitStrings(*_localizedStrings, gameToPromote, GetPromotionUrl(gameToPromote), GetPromotionUrl(GameInfo::getGameType()));
}

Win32NativeUiInterface::~Win32NativeUiInterface()
{
	if (!_secondInstance)
	{
		DeleteWCharVector(_localizedStrings);
		ReleaseMutex(_mutexHandle);
		CloseHandle(_mutexHandle);
	}

	delete _wrapper;
}

ConvertedSetting* Win32NativeUiInterface::CreateSetting(const BaseSetting* setting)
{
	ConvertedSetting* result;
	SettingControlType type = setting->getType();

	vector<wstring> v;

	switch (type)
	{
	case SettingControlType::RADIO_GROUP:
		result = new ConvertedRadioSetting();
		
		((ConvertedRadioSetting*)result)->ItemDescriptions = WStringToWCharVector(((RadioGroupSetting*)setting)->getItemDescriptions());		
		((ConvertedRadioSetting*)result)->ItemNames = WStringToWCharVector(((RadioGroupSetting*)setting)->getItemNames());

		break;
	case SettingControlType::SLIDER:
		result = new ConvertedSliderSetting();

		((ConvertedSliderSetting*)result)->Title = CreateWCharString(((SliderSetting*)setting)->getTitle());
		((ConvertedSliderSetting*)result)->MinValue = ((SliderSetting*)setting)->getMinValue();
		((ConvertedSliderSetting*)result)->MaxValue = ((SliderSetting*)setting)->getMaxValue();
		break;
	}

	result->Type = type;
	result->Value = setting->getIntValue();
	result->States = setting->getEnabledStates();

	return result;
}

void Win32NativeUiInterface::DeleteSetting(ConvertedSetting* setting)
{
	if (setting->Type == SettingControlType::RADIO_GROUP)
	{
		DeleteWCharVector(((ConvertedRadioSetting*)setting)->ItemDescriptions);
		DeleteWCharVector(((ConvertedRadioSetting*)setting)->ItemNames);
	}
	else
		delete ((ConvertedSliderSetting*)setting)->Title;

	delete setting;
}

void Win32NativeUiInterface::MaximizeGameWindow()
{
	if (isMidGame())
	{
		SDL_SysWMinfo info;
		SDL_VERSION(&info.version);
		SDL_GetWMInfo(&info);
		ShowWindow(info.window, SW_MAXIMIZE);
		_wrapper->SetOglWindowHandle(info.window);		
		SetActiveWindow(info.window);
		SetFocus(info.window);
	}
}

void Win32NativeUiInterface::MinimizeGameWindow()
{
	if (isMidGame())
	{
		SDL_SysWMinfo info;
		SDL_VERSION(&info.version);
		SDL_GetWMInfo(&info);
		ShowWindow(info.window, SW_HIDE);
	}
}

int Win32NativeUiInterface::GetStates()
{
	int menuState = 0;

	menuState |= isContinueEnabled() << 0;
	menuState |= isSaveEnabled() << 1;
	menuState |= isLoadEnabled() << 2;
	menuState |= hasTutorial() << 3;
	menuState |= isOpenHouseEnabled() << 4;
	menuState |= hasExtras() << 5;
	menuState |= hasCrossPromotionButton() << 6;

	return menuState;
}

void Win32NativeUiInterface::ShowMenuInternal(int menuAction)
{
	if (_localizedStrings == NULL)
	{		
		LoadStrings();
		
		std::string url = getCrossPromotionCheckUrl();
		std::wstring wideurl = std::wstring(url.begin(), url.end());
		const wchar_t* wideresponse;
		wideresponse = _wrapper->QueryUrl(wideurl.c_str());
		std::wstring w_response(wideresponse);
		std::string response(w_response.begin(), w_response.end());
		onCrossPromotionHttpResponse(response);	
		
	}

	MinimizeGameWindow();	

	do
	{
		_showingMenu = true;
		menuAction = _wrapper->ShowMenu(menuAction, menuAction == UnmanagedWrapper::MenuActionMain ? GetStates() : 0);

		switch (menuAction)
		{
		case UnmanagedWrapper::MenuActionSplash:
			menuAction = UnmanagedWrapper::MenuActionMain;
			break;
		case UnmanagedWrapper::MenuActionNewGame:
			onNewGamePressed();
			menuAction = UnmanagedWrapper::MenuActionMain;
			break;
		case UnmanagedWrapper::MenuActionContinue:
			onContinuePressed();
			break;
		case UnmanagedWrapper::MenuActionSave:
			onSavePressed();
			menuAction = UnmanagedWrapper::MenuActionMain;
			//_showingMenu = false;
			break;
		case UnmanagedWrapper::MenuActionLoad:
			onLoadPressed();
			menuAction = UnmanagedWrapper::MenuActionMain;
			//_showingMenu = !isMidGame();
			break;
		case UnmanagedWrapper::MenuActionSettings:
			onSettingsFromMainMenu();
			break;
		case UnmanagedWrapper::MenuActionSettingsVoice:
			ShowSettingsSubMenu(menuAction, SettingSubMenuType::VOICE);
			menuAction = UnmanagedWrapper::MenuActionSettings;
			break;
		case UnmanagedWrapper::MenuActionSettingsControls:
			ShowSettingsSubMenu(menuAction, SettingSubMenuType::CONTROLS);
			menuAction = UnmanagedWrapper::MenuActionSettings;
			break;
		case UnmanagedWrapper::MenuActionSettingsGraphics:
			ShowSettingsSubMenu(menuAction, SettingSubMenuType::GRAPHICS);
			menuAction = UnmanagedWrapper::MenuActionSettings;
			break;
		case UnmanagedWrapper::MenuActionSettingsLanguage:
			ShowSettingsSubMenu(menuAction, SettingSubMenuType::LANGUAGE);
			// reload localized strings
			DeleteWCharVector(_localizedStrings);
			LoadStrings();
			menuAction = UnmanagedWrapper::MenuActionSettings;
			break;
		case UnmanagedWrapper::MenuActionSettingsMusic:
			ShowSettingsSubMenu(menuAction, SettingSubMenuType::MUSIC);
			menuAction = UnmanagedWrapper::MenuActionSettings;
			break;
		/*case UnmanagedWrapper::MenuActionSettingsAbout:
			break;*/
		case UnmanagedWrapper::MenuActionOpenHouse:
			onOpenHousePressed();
			menuAction = UnmanagedWrapper::MenuActionOpenHouse;
			break;
		case UnmanagedWrapper::MenuActionQuit:
			onQuitPressed();
			break;
		case UnmanagedWrapper::MenuActionSettingsBack:
			onMainMenuFromSettings();
			menuAction = UnmanagedWrapper::MenuActionMain;
			break;
		case UnmanagedWrapper::MenuActionTutorial:
			showTutorialVideo();
			menuAction = UnmanagedWrapper::MenuActionMain;
			break;
		case UnmanagedWrapper::MenuActionEnableOpenHouse:
			onEnableOpenHouse();
			menuAction = UnmanagedWrapper::MenuActionMain;
			break;
		}
	} while (_showingMenu);
}

void Win32NativeUiInterface::showMenu(bool showSplash) 
{  
	/*
	if (_secondInstance)
	{
		LOGD("********** second instance detected, quitting");
		onQuitPressed();
		return;
	}
	*/
	MinimizeGameWindow();
	ShowMenuInternal(showSplash ? UnmanagedWrapper::MenuActionSplash : UnmanagedWrapper::MenuActionMain);
}

void Win32NativeUiInterface::ShowSettingsSubMenu(int menuAction, SettingSubMenuType type)
{
	int settingRetValue = 0;
	vector<ConvertedSetting*> convertedSettings;

	vector<const BaseSetting*> settings = getSettingsForType(type);
	onSpecificSettingsShown(type);

	do
	{
		// repack all setting
		for (vector<const BaseSetting*>::iterator iterator = settings.begin(); iterator != settings.end(); iterator++)
			convertedSettings.push_back(CreateSetting(*iterator));

		settingRetValue = _wrapper->ShowSettings(menuAction, convertedSettings);

		// update all settings
		for (int i = 0; i < convertedSettings.size(); i++)
			settings[i]->setIntValue(convertedSettings[i]->Value);

		// delete all created settings
		for (vector<ConvertedSetting*>::iterator iterator = convertedSettings.begin(); iterator != convertedSettings.end(); iterator++)
			delete *iterator;
		convertedSettings.clear();
	} while (settingRetValue != 0);
	
	settings.clear();
	onSpecificSettingsFinished(type);
}

void Win32NativeUiInterface::hideMenu() 
{
	_showingMenu = false;
	_wrapper->HideMenu();
	MaximizeGameWindow();
}

void Win32NativeUiInterface::showTutorialVideo()
{
	//LOGD("********** showTutorialVideo");
	ShowMenuInternal(UnmanagedWrapper::MenuActionTutorial);
}

void Win32NativeUiInterface::refreshSettingStates()
{
   
}

void Win32NativeUiInterface::BeforeDialog()
{
	if (!_showingMenu)
	{
		MinimizeGameWindow();
		_wrapper->DisplayMainMenu(GetStates());
	}
}

void Win32NativeUiInterface::AfterDialog()
{
	if (!_showingMenu)
	{
		hideMenu();
		MaximizeGameWindow();
	}
}

void Win32NativeUiInterface::showAlertDialog(wstring message, int buttonNum, wstring button1, wstring button2, wstring button3, 
	shared_ptr<Enhanced::AlertDialogListener> listener) 
{   
	BeforeDialog();
	
	// create an array of buttons
	vector<const wchar_t*> buttons;

	if (buttonNum > 0)
		buttons.push_back(button1.c_str());
	if (buttonNum > 1)
		buttons.push_back(button2.c_str());
	if (buttonNum > 2)
		buttons.push_back(button3.c_str());
	
	int result = _wrapper->ShowDialog(message.c_str(), buttons) - _wrapper->MenuActionAlertButton1;

	// if the lack of the call to AfterDialog() will cause bugs, try to uncomment it here
	//AfterDialog();

	listener->onClick(result);
}

void Win32NativeUiInterface::showInputDialog(wstring message, string initialInput, wstring negativeButton, wstring positiveButton, shared_ptr<Enhanced::InputDialogListener> listener) 
{	
	wstring wstringInput(initialInput.begin(), initialInput.end());
	wchar_t* wcharInput = (wchar_t*)wstringInput.c_str();
	
	int result = _wrapper->ShowDialogWithInput(message.c_str(), &wcharInput, negativeButton.c_str(), positiveButton.c_str());

	wstring wstringOutput(wcharInput);
	string stringOutput(wstringOutput.begin(), wstringOutput.end());
	
	listener->onResult(result == _wrapper->MenuActionSelectPositive, stringOutput);

	if (isMidGame() && (result != _wrapper->MenuActionSelectPositive))
		onContinuePressed();
}

void Win32NativeUiInterface::showListDialog(wstring message, vector<wstring> items, int buttonNum, wstring negativeButton, wstring positiveButton, shared_ptr<Enhanced::ListDialogListener> listener) 
{
	vector<const wchar_t*>* convertedItems = WStringToWCharVector(items); 
	
	BeforeDialog();

	int result = _wrapper->ShowDialogWithList(message.c_str(), *convertedItems, buttonNum, negativeButton.c_str(), positiveButton.c_str());

	DeleteWCharVector(convertedItems);

	listener->onResult(result != _wrapper->MenuActionSelectNegative, result - _wrapper->MenuActionListSlot1);

	AfterDialog();

	if (isMidGame() && (result == _wrapper->MenuActionSelectNegative))
	{
		onContinuePressed();
	}

	_showingMenu = !isMidGame() && (result == _wrapper->MenuActionSelectNegative);
}

void Win32NativeUiInterface::showProgressDialog(wstring message) 
{
	BeforeDialog();
	_wrapper->ShowProgressDialog(message.c_str());
}

void Win32NativeUiInterface::hideCurrentDialog() 
{	 
	_wrapper->HideProgressDialog();
	AfterDialog();
}

void Win32NativeUiInterface::openAppStoreForRating()
{

}

Enhanced::Locale Win32NativeUiInterface::getLocale() 
{
	ULONG numLanguages = 0;
	DWORD cchLanguagesBuffer = 0;
	BOOL hr = GetUserPreferredUILanguages(MUI_LANGUAGE_NAME, &numLanguages, NULL, &cchLanguagesBuffer);
	Locale locale = Locale::ENGLISH;

	if (hr) {
		WCHAR* pwszLanguagesBuffer = new WCHAR[cchLanguagesBuffer];
		hr = GetUserPreferredUILanguages(MUI_LANGUAGE_NAME, &numLanguages, pwszLanguagesBuffer, &cchLanguagesBuffer);
		if (hr) {
			if (wcsncmp(pwszLanguagesBuffer, L"es-", 3) == 0)
			{
				locale = Locale::SPANISH;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"pt-", 3) == 0)
			{
				locale = Locale::PORTUGESE;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"de-", 3) == 0)
			{
				locale = Locale::GERMAN;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"fr-", 3) == 0)
			{
				locale = Locale::FRENCH;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"cs-", 3) == 0)
			{
				locale = Locale::CZECH;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"he-", 3) == 0)
			{
				locale = Locale::HEBREW;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"it-", 3) == 0)
			{
				locale = Locale::ITALIAN;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"pl-", 3) == 0)
			{
				locale = Locale::POLISH;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"ru-", 3) == 0)
			{
				locale = Locale::RUSSIAN;
			}
			else if (wcsncmp(pwszLanguagesBuffer, L"sv-", 3) == 0)
			{
				locale = Locale::SWEDISH;
			}
			delete pwszLanguagesBuffer;
		}
	}

  	return locale;
}

void Win32NativeUiInterface::showSubtitles(wstring text) 
{
	/*wchar_t* message = CreateWCharString(text);
	_wrapper->ShowNotification(message);
	delete message;*/
}

void Win32NativeUiInterface::hideSubtitles() 
{   
	//_wrapper->HideMenu();
}

void Win32NativeUiInterface::showDeveloperLogo1() 
{
	LOGD("********** showDeveloperLogo1");
	MinimizeGameWindow();
	if (_wrapper->ShowMenu(UnmanagedWrapper::MenuActionDeveloperLogo1, 0) == UnmanagedWrapper::MenuActionQuit)
		onQuitPressed();
}
 
void Win32NativeUiInterface::showDeveloperLogo2() 
{
	LOGD("********** showDeveloperLogo2");
	MinimizeGameWindow();
	if (_wrapper->ShowMenu(UnmanagedWrapper::MenuActionDeveloperLogo2, 0) == UnmanagedWrapper::MenuActionQuit)
		onQuitPressed();
}

void Win32NativeUiInterface::hideDeveloperLogo() 
{
	LOGD("********** hideDeveloperLogo");
	hideMenu();
}

void Win32NativeUiInterface::sendGameStartAnalytics(SettingInfo& settingInfo)
{
    
}

void Win32NativeUiInterface::onAchievementReached(int achievementId)
{
     
}

bool Win32NativeUiInterface::hasTouchScreen()
{
	return _wrapper->HasTouchScreen();
}